#include "duktape.h"
#include "tcc-patch.h"
#include <stdio.h>
#include <string.h>

void handle_fatal(void *udata, const char *msg) {
  fprintf(stderr, "Fatal error in Javascript: %s\n",
          (msg ? msg : "no message"));
  exit(1);
}

void catch (void *ctx) {
  fprintf(stderr, "JavaScript reports the following error:\n\"%s\"\n",
          duk_safe_to_string(ctx, -1));
  duk_destroy_heap(ctx);
  exit(1);
}

int main(int argc, char **argv) {
  if (argc - 2 or strlen(argv[1]) < 5 or
      (strcmp(argv[1] + strlen(argv[1]) - 4, ".AEC") and
       strcmp(argv[1] + strlen(argv[1]) - 4, ".aec"))) {
    fprintf(stderr, "Please invoke this program from command-line and supply "
                    "an AEC source file as the only argument, for example:\n"
                    "Windows: aec.exe euclid.aec\n"
                    "Linux and MacOS: ./aec euclid.aec\n");
    return 1;
  }
#include "readCompiler.h"
  input = fopen("control.js", "r");
  if (!input) {
    fprintf(stderr, "Can't open 'control.js'?!\n");
    fflush(stderr);
    return 1;
  }
  while (!feof(input))
    if (fscanf(input, "%[^\n]\n", line)) {
      strcat(line, "\n");
      strcat(compilerString, line);
    }
  fclose(input);
  fprintf(stderr, "Initializing Duktape!\n");
  fflush(stderr);
  duk_context *ctx = duk_create_heap(0, 0, 0, 0, handle_fatal);
  duk_push_c_function(ctx, getIEEE754, 1);
  duk_put_global_string(ctx, "getIEEE754");
  duk_eval_string(ctx, compilerString);
  free(compilerString), free(line);
  duk_set_top(ctx, 1);
  fprintf(stderr, "Compilation started!\n");
  fflush(stderr);
  input = fopen(argv[1], "r");
  if (!input) {
    fprintf(stderr, "Can't read \'%s\'!\n", argv[1]);
    fflush(stderr);
    return 1;
  }
  int lineNumber = 0;
  line = calloc(512, 1);
  int hasError = 0;
  compilerString = calloc(1024, 1);
  if (!compilerString or !line) {
    fprintf(stderr, "Out of memory!?\n");
    return 1;
  }
  while (!feof(input)) {
    lineNumber++;
    if (fgets(line, 512, input)) {
      if (line[strlen(line) - 1] == '\n')
        line[strlen(line) - 1] = 0;
      if (line[strlen(line) - 1] == '\r')
        line[strlen(line) - 1] = 0;
      memset(compilerString, 0, 512);
      strcat(compilerString, "compileString(\"");
      int i;
      for (i = 0; line[i]; i++)
        if (line[i] == '\\')
          strcat(compilerString, "\\\\");
        else if (line[i] == '"')
          strcat(compilerString, "\\\"");
        else
          strncat(compilerString, line + i, 1);
      strcat(compilerString, "\")");
      duk_push_string(ctx, compilerString);
      if (duk_peval(ctx)) {
        const char *message = duk_safe_to_string(ctx, -1);
        if (strstr(message, "error"))
          hasError = 1;
        fprintf(stderr, "Message on line %d:\n\"%s\"\n", lineNumber, message);
      }
      duk_pop(ctx);
    }
  }
  free(line);
  free(compilerString);
  fclose(input);
  if (!hasError) {
    duk_eval_string(ctx, "stack.length");
    if (duk_get_int(ctx, -1)) {
      fprintf(stderr, "Control structures ('If', 'Else', 'While', 'EndIf' "
                      "and 'EndWhile') appear not to be properly nested, "
                      "compilation aborted!\n");
      fflush(stderr);
      return 1;
    }
    char *outputFileName = calloc(265, 1);
    strncat(outputFileName, argv[1], strlen(argv[1]) - 4);
    duk_eval_string(ctx, "syntax");
    const char *syntax = duk_get_string(ctx, -1);
    if (!strcmp(syntax, "fasm"))
      strcat(outputFileName, ".asm");
    else if (!strcmp(syntax, "gas"))
      strcat(outputFileName, ".s");
    else {
      fprintf(stderr,
              "JavaScript tried to open an invalid assembly language file "
              "(not ending in \".asm\" or \".s\") for output, quitting now!\n");
      duk_destroy_heap(ctx);
      return 1;
    }
    FILE *output = fopen(outputFileName, "w");
    free(outputFileName);
    if (!output) {
      fprintf(stderr, "Can't open output file!\n");
      fflush(stderr);
      return 1;
    }
    char commentSign = (strcmp(syntax, "fasm")) ? '#' : ';';
    duk_eval_string(ctx, "assembler");
    const char *assembler = duk_get_string(ctx, -1);
    fprintf(output,
            "%cGenerated by Arithmetic Expression Compiler "
            "(https://flatassembler.github.io/compiler.html) run in Duktape.\n"
            "%s",
            commentSign, assembler);
    fclose(output);
    fprintf(stderr, "Compilation finished without errors!\n");
  } else
    fprintf(stderr,
            "Compilation aborted due to the errors reported by JavaScript!\n");
  duk_destroy_heap(ctx);
  return 0;
}
